#gchord2voice.py

'''
The functions replaces a tune abc
with a new tune with the
guitar chords (gchords) replaced with a separate
voice where they have been replaced by notes.
For example the body:

%%MIDI gchord fzcz
|: cB | "F"A2F2 FAGF |
"C7"EGB2 BdcB |"F"Acde "Dm"fagf :|

is replaced with

V:1
|: cB | A2F2 FAGF |
EGB2 BdcB |Acde fagf :|
V:chords
 |: z2 |F,,z[F,A,C]z |
C,,z[C,E,G,B,]z |F,,z[D,F,A,]z :|

The guitar chords are expanded into accompaniment
using the same convention as abc2midi; however,
the gchord string cannot be arbitrary so that they
can be represented by musical symbols. (See abcguide.txt
which comes with the abcMIDI package for more details
regarding guitar chords.)

Implementation: the program makes a single pass thru the
tune separating the field commands from the body. If the
body is split into parts, each part is saved separately.

The K: L: and M: field commands contain the important
information which is used to set state variables (globals)
which determine how the notes in the body are processed.
Also the MIDI gchord string is extracted if present.

The body is processed line by line and note by note. As
the body is processed a new voice containing the expanded
guitar chords is built up and stored away.

Once the tune is finished, all the information is assembled
in an output tune which is returned.

'''


import re
import sys
import key
import meter
import gcgen
import gcstring
import gchord

#module key handles the K: command
#module meter handles the M:, L: command
#module gcstring handles the %%MIDI gchord string
#module gchord handles the specific guitar chord, eg. "Gm"
#module gcgen generates the expanded guitar chords using
#the information in gstring.

fieldpat = re.compile('X:|T:|C:|O:|A:|Q:|Z:|N:|H:|S:')
kfieldpat = re.compile('K:')
mfieldpat = re.compile('M:')
lfieldpat = re.compile('L:')
pfieldpat = re.compile('P:')
vfieldpat = re.compile('V:')
midicmdpat = re.compile('%%MIDI')
commentpat = re.compile('%')

line = ''
debug = False
preservegchord = False
xref = 'none'


def appendtext(line):
    global fieldtext,bodytext
    if body:
        bodytext = bodytext + line 
    else:
        fieldtext = fieldtext + line + '\n'


'''
To handle multiple parts, we save the body and chord voice
for each part  in separate strings which are put in
bodyvoicelist and chordvoicelist lists respectively.
'''

def initialize_g2v_globals ():
    global fieldtext, bodytext
    global bodyvoice, chordvoice, bodyvoicelist
    global chordvoicelist, partlabel
    global nparts
    global partlabel
    global expandedchords
    global nvoices
    global triplets_running
 
    bodytext = '' # saves the body belonging to a particular part
    bodyvoice = {} #references the body for each part detected
    chordvoice = {} #references the chordal accompaniment for a particular part
    bodyvoicelist = [] #a list of all the body parts
    chordvoicelist = [] #a list of all the chordal parts
    expandedchords = '' 
    fieldtext = '' #all the commands appearing before the first K:
    nparts = 0  # number of parts in the given tune
    nvoices = 0 # number of voices in the given tune
    partlabel = {}
    partlabel[0]='none'
    triplets_running = 0
    gchord.initialize_gchord()
    gcstring.initialize_gcstring()


def process_post_P(label):
    '''
    Handles P: field command found in the body of
    the abc file (i.e. after the K: field). The
    function saves the body text and the created chord
    voice for eventual output. The function is also
    called at the end of scanning any tune in order
    to transfer results to the bodyvoicelist and
    chordvoicelist.
    '''
    global bodyvoice,chordvoice,nparts
    global expandedchords,bodytext
    global bodyvoicelist,chordvoicelist
    # the P: command precedes the body
    if len(bodytext) < 2:
        partlabel[nparts] = label
        return
    bodyvoice[nparts] = bodytext
    chordvoice[nparts] = expandedchords
    bodyvoicelist.append(bodyvoice[nparts])
    chordvoicelist.append(chordvoice[nparts])
    nparts = nparts + 1
    partlabel[nparts] = label
    bodytext = ''
    expandedchords = ''




# the following two functions turn on or off the flag
# triplets_running. The flag is needed to determine the
# length of a particular note.
def reset_triplets():
    global triplets_running
    triplets_running = 0

def process_triplet():
    global triplets_running
    triplets_running = 1


# globals used by process_note
durpatf = re.compile('([0-9])\/([0-9])')
durpatn = re.compile('[0-9]')
durpatd = re.compile('(\/)([0-9])')
dur2    = re.compile('/+')


def process_chord(token):
    '''
    a chord eg [A3/2c3/2] is treated like a note
    except we ignore the time values of the other
    notes in the chord since they should all be
    the same.
    '''
    process_note(token)



def process_note(token):
    '''
    The main purpose of this function is to update
    the gcgen.bar_accumulator each time a note is
    processed.

    The function determines the time value of the note
    given in the token. Formats that are recognized are:
    C, C1/2 C/2 and C2. We ignore broken rhythms eg.
    C > D  and C < D because the time value of the two
    notes still add up to the same value. We ignore
    tied notes, since that also does not change the
    accumulated time of the scanned notes
    '''
    global triplets_running
    if durpatf.search(token):
        m = durpatf.search(token)
        num = m.group(1)
        denom = m.group(2)
        increment = meter.noteunits*int(num)/int(denom)
    elif durpatd.search(token):
        m = durpatd.search(token)
        denom = m.group(2)
        increment = meter.noteunits/int(denom)
    elif durpatn.search(token):
        m = durpatn.search(token)
        num = m.group()
        increment = meter.noteunits*int(num)
    elif dur2.search(token):
        m = dur2.search(token)
        increment = meter.noteunits/2
    else:
        increment = meter.noteunits

    if triplets_running:
        increment = 2*increment/3
        triplets_running = triplets_running+1
        if triplets_running > 3:
            triplets_running = 0

    gcgen.bar_accumulator = gcgen.bar_accumulator + increment



def process_gchord(token):
# separate the guitar chord from the double quotes
# determine the type of chord and find the equivalent
# notes.
    token = token.lstrip('"')
    token = token.rstrip('"')
    gchord.expandgchord(token)
    gcstring.interpret_gcstringlist()

def process_voice(line):
# extracts the voice number from a voice command
    global activevoice
    tline = line[2:]
    tline = tline.rsplit()
    activevoice = tline[0]




# globals used by process_line
barpat = re.compile('\|\[\d'
                    '|\|:'
                    '|:\|\[\d'
                    '|:\|'
                    '|::'
                    '|\|')
notepat = re.compile('(\^*|_*|=?)'
                     '([A-G]\,*|[a-g]\'*|z)'
                     '([0-9]*/*[0-9]*)'
                      '|/?[0-9]+')
gchordpat = re.compile('\"[^\"]+\"')
curlypat = re.compile('\{[^\}]*\}')
chordpat = re.compile('\[[^\]\[]*\]')
tripletpat = re.compile('\(3')
instructpat = re.compile('![^!]*!')


def process_line (line):
    '''
    Scans the notes in the body of the abc tune and
    processes the bar lines, guitar chord indications,
    triplet indications, chords, and notes as they
    are encountered going left to right. Tokens of
    these elements are identified as the scanner
    position moves incrementally.
    '''
    global bodytext
    global triplets_running
    global activevoice
    if preservegchord:
        bodytext = bodytext + line
    else:
        result = gchordpat.sub('',line)
        bodytext = bodytext + result + '\n'
    i = 0
    while i < len(line):
        m = barpat.match(line,i)        # bar line
        if m:
            i = m.span(0)[1]
            token = m.group(0)
            if activevoice == '1':
                gcgen.process_barline(token)
            key.resetkeytable()
            continue

        m = gchordpat.match(line,i)     #guitar chord
        if m:
            i = m.span(0)[1]
            token = m.group(0)
            process_gchord(token)
            continue

        m = curlypat.match(line,i)      #grace notes
        if m:
            i = m.span(0)[1]
            # skip grace notes
            continue

        m = tripletpat.match(line,i)    #triplet
        if m:
            i = m.span(0)[1]
            process_triplet()
            continue

        m = instructpat.match(line,i)   #instruction
        if m:
            i = m.span(0)[1]
            # skip instructions like !fff!
            continue

        m = notepat.match(line,i)       #note
        if m:
            i = m.span(0)[1]
            process_note(m.group(0))
            if activevoice == '1':
                gcgen.gchord_generator()
            continue

        m = chordpat.match(line,i)      #chord
        if m:
            i = m.span(0)[1]
            process_chord(m.group(0))
            continue


        i = i+1



import pdb



def g2v(abc):
    global body
    global expandedchords
    global line
    global nvoices
    global activevoice
    nvoices = 0
    activevoice = '1'
    initialize_g2v_globals()
    reset_triplets()
    body = False
    for line in abc.split('\n'):
        line = line.lstrip()
        if len(line) < 2 :
            break
        if debug:
            print('>>>',line)
        if fieldpat.match(line):        #ignore any field command except K: L: M: P: and V:
            appendtext(line)
        elif midicmdpat.match(line):    #%%MIDI command - ignore all except gchord
            appendtext(line)
            line = line.lstrip('%MIDI ')
            if line.find('gchord')==0:
                gcstring.process_gchord_cmd(line)
                gcgen.setup_gchord_generator()
        elif commentpat.match(line):    #ignore all other comments
            appendtext(line)
        elif kfieldpat.match(line):     #K: field get key signature and set tables
            appendtext(line)
            if body:
                expandedchords = expandedchords + line
            kfield = line.lstrip('K: ')
            sf     = key.key2sf(kfield)
            key.setupkey(sf)
            key.resetkeytable()
            body = True
        elif mfieldpat.match(line):     #M: grab meter and set up gchord generator
            appendtext(line)
            if body:
                expandedchords = expandedchords + line
            mfield = line.lstrip('M: ')
            mfield = mfield.rstrip()
            meter.setup_meter(mfield)
            gcgen.setup_gchord_generator()
        elif lfieldpat.match(line):     #L: commenad
            appendtext(line)
            lfield = line.lstrip('L: ')
            meter.setup_unitlength(lfield)
        elif pfieldpat.match(line):     #P: field
            process_post_P(line)
        elif vfieldpat.match(line):     #V: count voices
            process_voice(line)
            appendtext(line)
            nvoices = nvoices + 1
        elif body:              #body process music
            process_line(line)
            if len(gcgen.expandedchords_for_line) > 1:
                expandedchords = expandedchords + gcgen.expandedchords_for_line +'\n'
            gcgen.clear_expandedchords_for_line()

    # prepare processed tune for output
    gchordunitlength = gcgen.return_gchord_unitlength()

    process_post_P('none')
    output_text = fieldtext
    for i in range(0,nparts):
        if partlabel[i] != 'none':
            output_text = output_text + partlabel[i]
        output_text = output_text + 'V:1\n'
        output_text = output_text + bodyvoicelist[i]
        #print bodyvoicelist[i]
        output_text = output_text + 'V:chords clef=bass\n'
        output_text = output_text + gchordunitlength + '\n'
        output_text = output_text + chordvoicelist[i]

    return output_text


