#key.py
'''
The guitar chord pitches are expressed in MIDI pitch units. In
order to convert them to notes using midi2key we need to know whether they 
will appear with any accidentals. This depends upon the key signature.
The key signature given by the K: command is used to set up an
internal table keytable which does the conversion from MIDI pitch
units to notes. keytable is initialized from basetable at the
beginning of every bar. To permit propagation of accidentals,
keytable may be updated inside a bar.

To initialize basetable, we need to know the number of flats
or sharps in the key signature.
'''


import re
modeshift = {'maj':  0,  'min': -3, 'm': -3,   'aeo': -3,
            'loc': -5,  'ion': 0,  'dor': -2,  'phr': -4,
            'lyd': 1,   'mix': -1}


sharpnotes = {0: 'C',  1: '^C',  2: 'D',
              3: '^D', 4: 'E',   5: 'F',
              6: '^F', 7:  'G',  8: '^G',
              9: 'A',  10: '^A', 11: 'B' }

flatnotes  = {0: 'C',  1: '_D', 2: 'D',
              3: '_E', 4: 'E',  5: 'F',
              6: '_G', 7: 'G',  8: '_A',
              9: 'A', 10: '_B', 11: 'B' }


keypat =  re.compile('([A-G]|none)(b|#?)(.*)')
# for splitting a key signature into components (eg. F#dor into F # dor)

key = "FCGDAEB"

def key2sf(keysig):
# It interprets the key signature string in the K: command
# and determines the number of sharps/flats that are placed
# on the staff. For example Bb major has two flats so sf = -2
    m = keypat.match(keysig)
    if m:
        keycomponents = m.groups()
        tonic = keycomponents[0]
        accidental = keycomponents[1]
        mode = keycomponents[2]
        mode = mode.lower()
        sf = key.find(tonic)
        sf = sf - 1
        if accidental == '#':
            sf = sf + 7
        if accidental == 'b':
            sf = sf - 7
        if mode in modeshift:
            sf = sf + modeshift[mode]
    else:
        print('cannot understand', keysig)
    return sf

basekeytable = {}


def setupkey(sf):
# sets up table to convert MIDI pitches to
# notes taking into account the key signature
    global basekeytable
    if sf>=0:
        for i in range(12):
            basekeytable[i] = sharpnotes[i]
        if sf==0:
            return
        basekeytable[5] = '=F'
        basekeytable[6] = 'F'
        if sf==1:
            return
        basekeytable[0] = '=C'
        basekeytable[1] = 'C'
        if sf==2:
            return
        basekeytable[7] = '=G'
        basekeytable[8] = 'G'
        if sf==3:
            return
        basekeytable[2] = '=D'
        basekeytable[3] = 'D'
        if sf==4:
            return
        basekeytable[9] = '=A'
        basekeytable[10] = 'A'
        if sf==5:
            return
        basekeytable[4] = '=E'
        basekeytable[5] = 'E'
        if sf==6:
            return
        basekeytable[11] = '=B'
        basekeytable[0]  = 'B'
        return
    else:
        for i in range(0,12):
            basekeytable[i] = flatnotes[i]
        basekeytable[10] = 'B'
        basekeytable[11] = '=B'
        if sf==-1:
            return
        basekeytable[3] = 'E'
        basekeytable[4] = '=E'
        if sf==-2:
            return
        basekeytable[8] = 'A'
        basekeytable[9] = '=A'
        if sf==-3:
            return
        basekeytable[1] = 'D'
        basekeytable[2] = '=D'
        if sf==-4:
            return
        basekeytable[6] = 'G'
        basekeytable[7] = '=G'
        if sf==-5:
            return
        basekeytable[11] = 'C'
        basekeytable[0]  = '=C'
        if sf==-6:
            return
        basekeytable[4] = 'F'
        basekeytable[5] = '=F'
        return

keytable = {}

def resetkeytable():
    # copies basekeytable into keytable
    global keytable
    keytable.update(basekeytable)


def midi2key(midipitch):
#converts MIDI pitch to abc music notation note
    midi12 = midipitch % 12
    octave = midipitch /12
    note = keytable[midi12]
#propagate accidentals across the bar
    if note[0] == '^':
        keytable[midi12] = note[1]
        keytable[midi12-1] = '='+note[1]
    elif note[0] == '_':
        keytable[midi12] = note[1]
        keytable[midi12+1] = '='+note[1]
    elif note[0] == '=':
        keytable[midi12] = note[1]

    if octave==3: return note+',,'
    if octave==4: return note+','
    return note
